/**
 * @file Manages charging items. Charging items are items like dash or attack
 * that must be charged until a maximum level.
 *
 * All times are in seconds.
 */

/**
 * @typedef ChargingItem
 */

/**
 * @param {float} chargingLevelDuration  The time, in seconds, to chare one level.
 */
function BR_ChargingItem_create(levelChargingDuration) {
	let item = {
		// Should be constant.
		_levelChargingDuration : levelChargingDuration,

		// The current charging for going to the next level.
		_currentChargingDuration : 0,

		isCharging : false,
		level : 0,
	};
	return item;
}

/**
 * @param {ChargingItem} reset
 */
function BR_ChargingItem_reset(item) {
	item._currentChargingDuration = 0;
	item.isCharging = false;
	item.level = 0;
}

/**
 * @param {ChargingItem} reset
 * @param {float} elapsedTime  In seconds.
 * @param {unsigned int} maxLevel  The maximum level (included).
 */
function BR_ChargingItem_charge(item, elapsedTime, maxLevel) {
	if(item.isCharging) {
		item._currentChargingDuration += elapsedTime;
		let increaseLevel = (item._currentChargingDuration >= item._levelChargingDuration);
		if(increaseLevel) {
			item._currentChargingDuration -= item._levelChargingDuration;
			if(item.level < maxLevel) {
				++item.level;
			}
			if(item.level >= maxLevel) {
				// Maybe for some reason the new "maxLevel" is lower than the
				// previous "maxLevel". In that case, it is better to be sure
				// we are not over the new "maxLevel".
				item.level = maxLevel;

				// We want to start from 0 if a higher "maxLevel" is possible later.
				item._currentChargingDuration = 0;
			}
		}
	} else {
		BR_ChargingItem_reset(item);
		item.isCharging = true;
		// We do not add time because we just start charging right now.
	}
}
