/**
 * @file Defines the (nearly) constant configuration used in the project.
 */

// Game app configuration.
const BR_APP_WIDTH = 500;
const BR_APP_HEIGHT = 500;
const BR_APP_BRACKGROUND_COLOR = 0xB4D455;

// HTML elements.
const BR_CONTAINER_ELEMENT_ID = "BR-game-container";

// Paths.
const BR_PATH_ASSETS = "assets";
const BR_PATH_SPRITES = BR_PATH_ASSETS + "/sprites";
const BR_PATH_SPRITESHEET_INSECT = BR_PATH_SPRITES + "/insect.png";
const BR_PATH_SPRITESHEET_PLAYER = BR_PATH_SPRITES + "/player.png";

// Key configuration. The default is set here and should be modified by the
// player later.
// WARNING: "keyName" is the displayed name, not the "evt.key" from
// "document.onkeyup" or "document.onkeydown".
const BR_KEYS_CONFIGURATION = {
	confirm : { keyCode: 13, keyName: "Enter" },
	cancel : { keyCode: 27, keyName: "Escape" },
	up : { keyCode: 38, keyName: "Up" },
	down : { keyCode: 40, keyName: "Down" },
	left : { keyCode: 37, keyName: "Left" },
	right : { keyCode: 39, keyName: "Right" },
	dash : { keyCode: 88, keyName: "x" },
	attack : { keyCode: 67, keyName: "c" },
};

// Enum state for the key.
const BR_KEY_STATE = {
	UP : 0,
	// JUST_DOWN : 1,
	DOWN : 2,
	// JUST_UP : 4,
};

// The different screen the player interacts with..
const BR_SCREEN_STATE = {
	MAIN_MENU : 1,
	GAME : 2,
	GAME_OVER : 3,
};

const BR_PLAYER_LIFE = 3;
const BR_ENEMY_LIFE = 1;

// When sprites are drawn, the head is not necessarily to the right;
// so this means the character does not start with an angle of 0.
// Angle here is in radian.
const BR_SPRITE_DEFAULT_ANGLE = -Math.PI / 2;

// Squared insect hitboax radius.
const BR_INSECT_HITBOX_RADIUS_2 = Math.pow(10, 2);

// Speed configuration.
// In PIXI.Sprite: use "angle" (dregrees), not "rotation" (radian).
// Also be careful, PIXIjs does rotation clockwise; so mathematically inverted.
// (Maybe because Y axis is inverted.)
const BR_ANGLE_IN_DEGREES_BY_SECONDS = 240;
const BR_SPEED_FORWARD_BY_SECONDS = 200;
const BR_SPEED_BACKWARD_BY_SECONDS = 30;
const BR_SPEED_FORWARD_WHILE_CHARGING_BY_SECONDS = 25;  // While charging dash or attack.
const BR_SPEED_BACKWARD_WHILE_CHARGING_BY_SECONDS = 20;  // While charging dash or attack.

// Dash management. Time is in seconds.
const BR_DASH_SPEED_BY_SECONDS = 600;
const BR_DASH_LEVEL_DURATIONS = [0, 0.25, 0.5, 1];
const BR_DASH_CHARGING_LEVEL_DURATION = 0.5;

// Attack management.
const BR_ATTACK_CHARGING_LEVEL_DURATION = 0.5;
const BR_ATTACK_DURATION = 1;
const BR_ATTACK_MAX_LEVEL = 3;

// Bottle.
const BR_BOTTLE_WIDTH = 200;
const BR_BOTTLE_HEIGHT = 200;
const BR_BOTTLE_BORDER_THICKNESS = 3;
const BR_BOTTLE_BORDER_COLOR = 0x000000;
const BR_BOTTLE_WALL_THICKNESS = 32;  // In pixels.
const BR_BOTTLE_DASH_LEVEL_DURATIONS = [0, 0.25, 0.5, 1];
const BR_BOTTLE_DASH_SPEED = 500;
const BR_BOTTLE_COLOR_WALL = 0x729FCF;
const BR_BOTTLE_COLOR_FLOOR = BR_APP_BRACKGROUND_COLOR;

// Avoid being stuck against the wall.
const BR_MARGIN_INSECT_WALL = 1;
const BR_MARGIN_BOTTLE = 1;

// Font configuration.
const BR_TEXT_NORMAL_OPTIONS = {
	fontFamily: 'Arial',
	fontSize: 20,
	// fontStyle: 'italic',
	fontWeight: 'bold',
	// fill: ['#ffffff', '#00ff99'], // gradient
	fill: ["#ffffff" ],
	stroke: '#4a1850',
	strokeThickness: 2,
	// dropShadow: true,
	// dropShadowColor: '#000000',
	// dropShadowBlur: 4,
	// dropShadowAngle: Math.PI / 6,
	// dropShadowDistance: 6,
	// wordWrap: true,
	// wordWrapWidth: 440,
	// lineJoin: 'round',
};

const BR_TEXT_GAME_TITLE_OPTIONS = {
	fontFamily: 'Arial',
	fontSize: 64,
	fontStyle: 'italic',
	fontWeight: 'bold',
	fill: ['#729FCF', '#B58D4C'], // gradient
	stroke: '#4a1850',
	strokeThickness: 2,
	dropShadow: true,
	dropShadowColor: '#000000',
	dropShadowBlur: 4,
	dropShadowAngle: Math.PI / 6,
	dropShadowDistance: 6,
	wordWrap: true,
	wordWrapWidth: 440,
	lineJoin: 'round',
};

const BR_TEXT_CREDITS_OPTIONS = {
	fontFamily: 'Arial',
	fontSize: 20,
	// fontStyle: 'italic',
	fontWeight: 'bold',
	// fill: ['#ffffff', '#00ff99'], // gradient
	fill: ["#ffffff" ],
	stroke: '#4a1850',
	strokeThickness: 2,
	// dropShadow: true,
	// dropShadowColor: '#000000',
	// dropShadowBlur: 4,
	// dropShadowAngle: Math.PI / 6,
	// dropShadowDistance: 6,
	// wordWrap: true,
	// wordWrapWidth: 440,
	// lineJoin: 'round',
};
