/**
 * @file Game starting point.
 */

/**
 * @typedef Enemy
 * @property {Insect} insect
 * @property {object} ai  Handle the A.I. of the enemy.
 */

var BR_mainMenuState = {
};

var BR_gameOverState = {
	text : null,
	textStyle : null,
};

var BR_gameState = {
	/** @type Insect */
	player : null,

	/** @typedef [Enemy...] */
	enemies : [],

	// It should not be necessary but I have a bug if I try to reuse dead insects.
	// Necromancy does not work.
	enemiesToRemove : [],

	/** @type Bottle */
	bottle : null,

	gameTime : 0,

	timeLabel : null,
	gameTimeGraphics : null,

	textStyle : null,
};

function BR_createEnemyAI() {
	let ai = {
		remainingAngle : 0,
	};
	return ai
}

function BR_createEnemy() {
	let stats = {
		forwardSpeed : BR_SPEED_FORWARD_BY_SECONDS,
		backwardSpeed : BR_SPEED_BACKWARD_BY_SECONDS,
		forwardSpeedWhileCharging : BR_SPEED_FORWARD_WHILE_CHARGING_BY_SECONDS,
		backwardSpeedWhileCharging : BR_SPEED_BACKWARD_WHILE_CHARGING_BY_SECONDS,
		dashSpeed : BR_DASH_SPEED_BY_SECONDS,
		angleSpeed : BR_SPEED_FORWARD_BY_SECONDS,
		hitboxRadius2 : BR_INSECT_HITBOX_RADIUS_2,
	};
	let sprite = BR_createInsectSprite();
	let dashDuration = BR_DASH_CHARGING_LEVEL_DURATION;
	let attackDuration = BR_ATTACK_CHARGING_LEVEL_DURATION;
	let life = BR_ENEMY_LIFE;
	let enemy = {
		insect : BR_Insect_create(stats, sprite, dashDuration, attackDuration, life),
		ai : BR_createEnemyAI(),
	};
	return enemy;
}

function BR_onkeydown(evt) {
	switch(evt.keyCode) {
		case BR_KEYS_CONFIGURATION.confirm.keyCode:
			BR_keys_state.confirm = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.cancel.keyCode:
			BR_keys_state.cancel = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.up.keyCode:
			BR_keys_state.up = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.down.keyCode:
			BR_keys_state.down = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.left.keyCode:
			BR_keys_state.left = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.right.keyCode:
			BR_keys_state.right = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.dash.keyCode:
			BR_keys_state.dash = BR_KEY_STATE.DOWN;
			break;
		case BR_KEYS_CONFIGURATION.attack.keyCode:
			BR_keys_state.attack = BR_KEY_STATE.DOWN;
			break;
		default:
			return;
	}
}

function BR_onkeyup(evt) {
	switch(evt.keyCode) {
		case BR_KEYS_CONFIGURATION.confirm.keyCode:
			BR_keys_state.confirm = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.cancel.keyCode:
			BR_keys_state.cancel = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.up.keyCode:
			BR_keys_state.up = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.down.keyCode:
			BR_keys_state.down = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.left.keyCode:
			BR_keys_state.left = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.right.keyCode:
			BR_keys_state.right = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.dash.keyCode:
			BR_keys_state.dash = BR_KEY_STATE.UP;
			break;
		case BR_KEYS_CONFIGURATION.attack.keyCode:
			BR_keys_state.attack = BR_KEY_STATE.UP;
			break;
		default:
			return;
	}
}

function BR_updateGlobalTime() {
	let now = Date.now() / 1000;
	BR_elapsedTime = now - BR_lastTimeUpdate;
	BR_lastTimeUpdate = now;
}

function BR_shouldInsectDash(gameState, insect) {
	// TODO: improve
	let chance = 0.001;
	let r = Math.random();
	let result = r < chance;
	return result;
}

function BR_chargeInsectDashUntil(insect, level, elapsedTime) {
	if(insect.dashCharging.level == level) {
		return;
	}
	let maxLevel = BR_DASH_LEVEL_DURATIONS.length - 1;
	if(level < maxLevel) {
		maxLevel = level;
	}
	BR_ChargingItem_charge(insect.dashCharging, elapsedTime, maxLevel);
}

function BR_isInsectHurtByBottle(insect, bottle, keepInBottleResult) {
	let hasHitWall = keepInBottleResult.left;
	hasHitWall ||= keepInBottleResult.right;
	hasHitWall ||= keepInBottleResult.up;
	hasHitWall ||= keepInBottleResult.down;
	if(!hasHitWall) {
		return false;
	}
	if(!BR_Bottle_isDashing(bottle)) {
		return false;
	}

	let hurt = false
	if(keepInBottleResult.left && bottle.dashOrientation.x > 0) {
		hurt = true;
	}
	if(keepInBottleResult.right && bottle.dashOrientation.x < 0) {
		hurt = true;
	}
	if(keepInBottleResult.up && bottle.dashOrientation.y > 0) {
		hurt = true;
	}
	if(keepInBottleResult.down && bottle.dashOrientation.y < 0) {
		hurt = true;
	}

	return hurt;
}

function BR_addEnemy(gameState, position) {
	let enemy = BR_createEnemy();
	let bottle = BR_gameState.bottle;
	enemy.insect.sprite.x = bottle.position.x + bottle.width / 2;
	enemy.insect.sprite.y = bottle.position.y + bottle.height / 2;
	BR_app.stage.addChild(enemy.insect.sprite);
	gameState.enemies.push(enemy);
}

function BR_managePlayer(elapsedTime, gameState) {
	let player = gameState.player;

	// When dashing or attacking, the player cannot chose anymore its direction.
	if(BR_Insect_isDashing(player)) {
		BR_Insect_dash(player, elapsedTime);
	}  else {
		// Move.
		if(BR_keys_state.left == BR_KEY_STATE.DOWN) {
			let angle = BR_ANGLE_IN_DEGREES_BY_SECONDS * elapsedTime;
			player.sprite.angle -= angle;
		}
		if(BR_keys_state.right == BR_KEY_STATE.DOWN) {
			let angle = BR_ANGLE_IN_DEGREES_BY_SECONDS * elapsedTime;
			player.sprite.angle += angle;
		}
		if(BR_keys_state.up == BR_KEY_STATE.DOWN) {
			BR_Insect_moveForward(player, elapsedTime);
		}
		if(BR_keys_state.down == BR_KEY_STATE.DOWN) {
			BR_Insect_moveBackward(player, elapsedTime);
		}
	}

	// Dash.
	if(!BR_Insect_isDashing(player) && !player.attackState.doing) {
		if(BR_keys_state.dash == BR_KEY_STATE.DOWN) {
			// Charging the dash.
			let previousLevel = player.dashCharging.level;
			let maxLevel = BR_DASH_LEVEL_DURATIONS.length - 1;
			BR_ChargingItem_charge(player.dashCharging, elapsedTime, maxLevel);
			if(previousLevel < player.dashCharging.level) {
				console.log("New dash level: " + player.dashCharging.level);
			}
		} else {
			if(player.dashCharging.isCharging) {
				// Start to dash.
				// TODO: use BR_Insect_startDash
				let level = player.dashCharging.level;
				if(level > 0) {
					let remainingTime = BR_DASH_LEVEL_DURATIONS[level];
					BR_ItemState_start(player.dashState, remainingTime)
				}
				BR_ChargingItem_reset(player.dashCharging);
			}
		}
	}

	// // Attack.
	// if(!BR_Insect_isDashing(player) && !player.attackState.doing) {
	// 	if(BR_keys_state.attack == BR_KEY_STATE.DOWN) {
	// 		let previousLevel = player.attackCharging.level;
	// 		let maxLevel = BR_ATTACK_MAX_LEVEL;
	// 		BR_ChargingItem_charge(player.attackCharging, elapsedTime, maxLevel);
	// 		if(previousLevel < player.attackCharging.level) {
	// 			console.log("New attack level: " + player.attackCharging.level);
	// 		}
	// 	} else {
	// 		if(player.attackCharging.isCharging) {
	// 			let level = player.attackCharging.level;
	// 			if(level > 0) {
	// 				let remainingTime = BR_ATTACK_DURATION;
	// 				BR_ItemState_start(player.attackState, remainingTime);
	// 			}
	// 			BR_ChargingItem_reset(player.attackCharging);
	// 		}
	// 	}
	// }

	// Hit enemy if dashing.
	if(BR_Insect_isDashing(player)) {
		for(let enemy of gameState.enemies) {
			let enemyInsect = enemy.insect;
			let position1 = player.sprite.position;
			let radius1 = BR_INSECT_HITBOX_RADIUS_2;
			let position2 = enemyInsect.sprite.position;
			let radius2 = BR_INSECT_HITBOX_RADIUS_2;
			if(BR_intersectsHitboxes(position1, radius1, position2, radius2)) {
				BR_Insect_getHit(enemyInsect, 1);
			}
		}
	}

	// Keep the player in the bottle.
	{
		let insect = player;
		let bottle = gameState.bottle;

		let result = BR_Insect_keepInBottle(insect, bottle);

		let hasHitWall = result.left || result.right || result.up || result.down;
		if(hasHitWall) {
			if(BR_Insect_isDashing(insect)) {
				BR_Insect_stopDash(insect);
				// TODO: improve level choice based on dashRemaining time.
				let level = 2;
				BR_Bottle_startDash(bottle, result.left, result.right, result.up, result.down, level);
			} else {
				if(BR_isInsectHurtByBottle(insect, bottle, result)) {
					BR_Insect_getHit(insect, 1);
				}
			}
		}
	}
}

function BR_manageEnemy(elapsedTime, gameState, enemy) {
	let insect = enemy.insect;

	if(BR_Insect_dead(insect)) {
		BR_app.stage.removeChild(insect.sprite);
		gameState.enemiesToRemove.push(enemy);
		return;
	}

	if(BR_Insect_isDashing(insect)) {
		BR_Insect_dash(insect, elapsedTime);
	} else if(BR_Insect_isChargingDash(insect) || BR_shouldInsectDash(insect)) {
		let level = 2;  // TOD0: chose a random level
		BR_chargeInsectDashUntil(insect, level, elapsedTime);
		if(insect.dashCharging.level == level) {
			BR_Insect_startDash(insect);
		}
	} else {
		// Randomly move the insect.
		// Chose a new random angle to turn if necessary.
		if(enemy.ai.remainingAngle == 0) {
			let angle = Math.random() * 180;
			if(Math.random() >= 0.5) {
				angle = -angle;
			}
			enemy.ai.remainingAngle = angle;
		}

		// Rotate as much as possible.
		let angle = BR_Insect_rotate(insect, enemy.ai.remainingAngle, elapsedTime);
		if(Math.abs(enemy.ai.remainingAngle) > Math.abs(angle)) {
			enemy.ai.remainingAngle -= angle;
		} else {
			enemy.ai.remainingAngle = 0;
		}

		// You're lonely rolling star, さあ前を向いて行こう!
		BR_Insect_moveForward(insect, elapsedTime);
	}

	// Keep the insect in the bottle.
	{
		let bottle = gameState.bottle;

		let result = BR_Insect_keepInBottle(insect, bottle);
		let hasHitWall = result.left || result.right || result.up || result.down;
		if(hasHitWall) {
			if(BR_Insect_isDashing(insect)) {
				BR_Insect_stopDash(insect);
				// TODO: improve level choice based on dashRemaining time.
				let level = 2;
				BR_Bottle_startDash(bottle, result.left, result.right, result.up, result.down, level);
			} else {
				if(BR_isInsectHurtByBottle(insect, bottle, result)) {
					BR_Insect_getHit(insect, 1);
				}
			}
		}
	}

	if(BR_Insect_dead(insect)) {
		BR_app.stage.removeChild(insect.sprite);
		gameState.enemiesToRemove.push(enemy);
	}
}

function BR_manageBottle(elapsedTime, gameState) {
	let bottle = gameState.bottle;

	// Dash.
	if(BR_Bottle_isDashing(bottle)) {
		BR_Bottle_dash(bottle, elapsedTime);
	}

	BR_bottle_updateGraphicsZone(bottle);
}

function BR_shouldAddEnemy(gameState) {
	let chance = 0.01;
	let r = Math.random();
	let result = r < chance;
	return result;
}

function BR_switchToScreen(newScreen) {
	BR_currentScreen = newScreen;
}

// To do when it is not game over.
function BR_doGameScreen() {
	// Update game time and its displayed label.
	{
		BR_gameState.gameTime += BR_elapsedTime;
		let text = "" + Math.floor(BR_gameState.gameTime);
		BR_app.stage.removeChild(BR_gameState.gameTimeGraphics);
		BR_gameState.gameTimeGraphics = new PIXI.Text(text, BR_gameState.textStyle);
		BR_gameState.gameTimeGraphics.x = BR_gameState.timeLabel.x + BR_gameState.timeLabel.width + 5;
		BR_gameState.gameTimeGraphics.y = BR_gameState.timeLabel.y;
		BR_app.stage.addChild(BR_gameState.gameTimeGraphics);
	}

	if(BR_shouldAddEnemy(BR_gameState)) {
		let x = BR_app.screen.width / 2;
		let y = BR_app.screen.height / 2;
		BR_addEnemy(BR_gameState, { x : x, y : y });
	}

	BR_manageBottle(BR_elapsedTime, BR_gameState);
	BR_managePlayer(BR_elapsedTime, BR_gameState);
	for(let enemy of BR_gameState.enemies) {
		BR_manageEnemy(BR_elapsedTime, BR_gameState, enemy);
	}

	for(let enemy of BR_gameState.enemiesToRemove) {
		let index = BR_gameState.enemies.indexOf(enemy);
		if(index > -1) {
		  BR_gameState.enemies.splice(index, 1);
		}
	}

	if(BR_Insect_dead(BR_gameState.player)) {
		BR_switchToScreen(BR_SCREEN_STATE.GAME_OVER);
	}
}

function BR_initializeMainMenu() {
	const normalStyle = new PIXI.TextStyle(BR_TEXT_NORMAL_OPTIONS);
	const titleStyle = new PIXI.TextStyle(BR_TEXT_GAME_TITLE_OPTIONS);
	const creditsStyle = new PIXI.TextStyle(BR_TEXT_CREDITS_OPTIONS);

	const margin = 10;

	let titleText = "Bottle Royale";
	let title = new PIXI.Text(titleText, titleStyle);
	BR_centerGraphicsOnScreen(title);
	BR_app.stage.addChild(title);

	let text = "In game, press [X] to charge dash.";
	text += "\nBe careful, other insects can dash too!";
	let instructions = new PIXI.Text(text, normalStyle);
	instructions.x = margin;
	instructions.y = title.y + title.height + 50;
	BR_app.stage.addChild(instructions);

	let creditsText = "Game by Nales, for Ludum Dare 54 (Limited Space)"
	let credits = new PIXI.Text(creditsText, creditsStyle);
	credits.x = BR_app.screen.width - credits.width - margin;
	credits.y = BR_app.screen.height - credits.height - margin;
	BR_app.stage.addChild(credits);
}

function BR_doMainMenuScreen() {
	if(BR_keys_state.confirm == BR_KEY_STATE.DOWN || BR_keys_state.dash == BR_KEY_STATE.DOWN) {
		BR_switchToScreen(BR_SCREEN_STATE.GAME);
	}
}

function BR_initializeGameOverScreen(gameTime) {
	BR_mainMenuState.textStyle = new PIXI.TextStyle(BR_TEXT_NORMAL_OPTIONS);

	let text = "You were crushed!";
	text += "\n";
	text += "\nYou survived " + Math.floor(gameTime) + " seconds.";
	text += "\n";
	text += "\nPress [Enter] to retry.";
	text += "\nPress [Escape] to go back to main menu.";

	BR_mainMenuState.text = new PIXI.Text(text, BR_gameState.textStyle);
	BR_centerGraphicsOnScreen(BR_mainMenuState.text);
	BR_app.stage.addChild(BR_mainMenuState.text);
}

function BR_doGameOverScreen() {
	if(BR_keys_state.confirm == BR_KEY_STATE.DOWN) {
		BR_switchToScreen(BR_SCREEN_STATE.GAME);
	}
	if(BR_keys_state.cancel == BR_KEY_STATE.DOWN) {
		BR_switchToScreen(BR_SCREEN_STATE.MAIN_MENU);
	}
}

// The method takes as a parameter a "delta" from "BR_app.ticker".
// I have no idea what this "delta" means so I do not rely on it.
function BR_doGameLoop(_) {
	BR_updateGlobalTime();

	// Initialize the new screen.
	if(BR_previousScreen != BR_currentScreen) {
		BR_app.stage.removeChildren();
		switch(BR_currentScreen) {
			case BR_SCREEN_STATE.MAIN_MENU:
				BR_initializeMainMenu();
				break;
			case BR_SCREEN_STATE.GAME:
				BG_initialize_gameState(BR_gameState);
				break;
			case BR_SCREEN_STATE.GAME_OVER:
				BR_initializeGameOverScreen(BR_gameState.gameTime);
				break;
		}
	}

	BR_previousScreen = BR_currentScreen;

	// Continue the current screen.
	switch(BR_currentScreen) {
		case BR_SCREEN_STATE.MAIN_MENU:
			BR_doMainMenuScreen();
			break;
		case BR_SCREEN_STATE.GAME:
			BR_doGameScreen();
			break;
		case BR_SCREEN_STATE.GAME_OVER:
			BR_doGameOverScreen();
			break;
	}
}

function BR_createPlayerSprite() {
	let sprite = PIXI.Sprite.from(BR_PATH_SPRITESHEET_PLAYER);
	sprite.anchor.set(0.5);
	return sprite;
}

function BR_createInsectSprite() {
	let sprite = PIXI.Sprite.from(BR_PATH_SPRITESHEET_INSECT);
	sprite.anchor.set(0.5);
	return sprite;
}

function BR_centerGraphicsOnScreen(graphics) {
	graphics.position.x = (BR_app.screen.width - graphics.width) / 2;
	graphics.position.y = (BR_app.screen.height - graphics.height) / 2;
}

function BG_initialize_gameState() {
	// Create bottle.
	{
		let width = BR_BOTTLE_WIDTH;
		let height = BR_BOTTLE_HEIGHT;
		let position = {
			x : (BR_app.screen.width - width) / 2,
			y : (BR_app.screen.height - height) / 2,
		};
		let boundsMargin = 10;
		let bounds = {
			x : boundsMargin,
			y : boundsMargin,
			width : BR_app.screen.width - 2 * boundsMargin,
			height : BR_app.screen.height - 2 * boundsMargin,
		}
		BR_gameState.bottle = BR_Bottle_create(position, width, height, bounds);
		BR_app.stage.addChild(BR_gameState.bottle.graphicsZone);
	}

	// Create player.
	{
		let stats = {
			forwardSpeed : BR_SPEED_FORWARD_BY_SECONDS,
			backwardSpeed : BR_SPEED_BACKWARD_BY_SECONDS,
			forwardSpeedWhileCharging : BR_SPEED_FORWARD_WHILE_CHARGING_BY_SECONDS,
			backwardSpeedWhileCharging : BR_SPEED_BACKWARD_WHILE_CHARGING_BY_SECONDS,
			dashSpeed : BR_DASH_SPEED_BY_SECONDS,
			angleSpeed : BR_ANGLE_IN_DEGREES_BY_SECONDS,
			hitboxRadius2 : BR_INSECT_HITBOX_RADIUS_2,
		};
		let sprite = BR_createPlayerSprite();
		let dashDuration = BR_DASH_CHARGING_LEVEL_DURATION;
		let attackDuration = BR_ATTACK_CHARGING_LEVEL_DURATION;
		let life = BR_PLAYER_LIFE;
		BR_gameState.player = BR_Insect_create(stats, sprite, dashDuration, attackDuration, life);

		BR_gameState.player.sprite.x = BR_app.screen.width / 2;
		BR_gameState.player.sprite.y = BR_app.screen.height / 2;
		BR_app.stage.addChild(BR_gameState.player.sprite);
	}

	// Create enemy.
	{
		let x = BR_app.screen.width / 2;
		let y = BR_app.screen.height / 2;
		BR_addEnemy(BR_gameState, { x : x, y : y });
	}

	BR_gameState.textStyle = new PIXI.TextStyle(BR_TEXT_NORMAL_OPTIONS);

	// Create time graphics.
	{
		BR_gameState.timeLabel = new PIXI.Text("Time:", BR_gameState.textStyle);
		BR_gameState.timeLabel.x = 5;
		BR_gameState.timeLabel.y = 5;
		BR_app.stage.addChild(BR_gameState.timeLabel);

		BR_gameState.gameTime = 0;
		BR_gameState.gameTimeGraphics = new PIXI.Text("0", BR_gameState.textStyle);
		BR_gameState.gameTimeGraphics.x = BR_gameState.timeLabel.x + BR_gameState.timeLabel.width + 5;
		BR_gameState.gameTimeGraphics.y = BR_gameState.timeLabel.y;
		BR_app.stage.addChild(BR_gameState.gameTimeGraphics);
	}
}

function BR_main() {
	document.onkeydown = BR_onkeydown;
	document.onkeyup = BR_onkeyup;

	// Create the game view.
	let appOptions = {
		width: BR_APP_WIDTH,
		height: BR_APP_HEIGHT,
		backgroundColor: BR_APP_BRACKGROUND_COLOR
	};
	BR_app = new PIXI.Application(appOptions);
	let container = document.getElementById(BR_CONTAINER_ELEMENT_ID);
	container.appendChild(BR_app.view);

	// Add game loop call every ticker time.
	BR_app.ticker.add(BR_doGameLoop);
}

window.onload = BR_main;
