/**
 * @file Various functions.
 */

/**
 * @typedef Vector2D
 * @property {float} x
 * @property {float} y
 */

/**
 * Returns the orientation the sprite is going forward.
 * @return {Vector2D}
 */
function BR_getNormalOrientation(sprite) {
		let currentAngle = sprite.rotation + BR_SPRITE_DEFAULT_ANGLE;
		let x = Math.cos(currentAngle);
		let y = Math.sin(currentAngle);
	let orientation = { x : x, y : y };
	return orientation;
}

function BR_getDistance2(point1, point2) {
	let dx = point1.x - point2.x;
	let dy = point1.y - point2.y;
	let distance = dx * dx + dy * dy;
	return distance;
}

// radius is squared.
function BR_intersectsHitboxes(position1, radius1, position2, radius2) {
	let distance = BR_getDistance2(position1, position2);
	let intersects = distance < (radius1 + radius2);
	return intersects;
}
function BR_createCircle(position, radius, color) {
	const circle = new PIXI.Graphics();
	circle.lineStyle(0); // draw a circle, set the lineStyle to zero so the circle doesn't have an outline
	circle.beginFill(color, 1);
	circle.drawCircle(0, 0, radius);
	circle.endFill();
	circle.position.x = position.x;
	circle.position.y = position.y;
	return circle;
}

function BR_createRectangle(position, width, height, color) {
	const rectangle = new PIXI.Graphics();
	rectangle.beginFill(color);
	rectangle.drawRect(0, 0, width, height);
	rectangle.endFill();
	rectangle.position.x = position.x
	rectangle.position.y = position.y
	return rectangle;
}

function BR_createRectangleWithBorder(position, insideWidth, insideHeight, insideColor, borderThickness, borderColor) {
	const rectangle = new PIXI.Graphics();
	// Draw the borders and fill the inside.
	rectangle.beginFill(borderColor);
	rectangle.drawRect(0, 0, insideWidth + 2 * borderThickness, insideHeight + 2 * borderThickness);
	rectangle.endFill();
	// Draw the inside.
	rectangle.beginFill(insideColor);
	rectangle.drawRect(borderThickness, borderThickness, insideWidth, insideHeight);
	rectangle.endFill();
	rectangle.position.x = position.x
	rectangle.position.y = position.y
	return rectangle;
}

function BR_normalize(vector) {
	var length2 = vector.x * vector.x + vector.y * vector.y;
	if(length2 == 0) {
		return { x : 0, y : 0 };
	}
	var length = Math.sqrt(length2);
	var normal = { x : vector.x / length, y : vector.y / length };
	return normal;
}
